@@ -24,8 +24,7 @@
 
 #include "ioctl.h"
 #include "util.h"
-
-static bool nvme_debug;
+#include "log.h"
 
 static int nvme_verify_chr(int fd)
 {
@@ -79,9 +78,10 @@ int nvme_get_nsid(int fd, __u32 *nsid)
 	return -1 * (errno != 0);
 }
 
-static int nvme_submit_passthru64(int fd, unsigned long ioctl_cmd,
-				  struct nvme_passthru_cmd64 *cmd,
-				  __u64 *result)
+__attribute__((weak))
+int nvme_submit_passthru64(int fd, unsigned long ioctl_cmd,
+			   struct nvme_passthru_cmd64 *cmd,
+			   __u64 *result)
 {
 	int err = ioctl(fd, ioctl_cmd, cmd);
 
@@ -90,62 +90,14 @@ static int nvme_submit_passthru64(int fd, unsigned long ioctl_cmd,
 	return err;
 }
 
-static void nvme_show_command(struct nvme_passthru_cmd *cmd, int err, struct timeval start,
-			      struct timeval end)
+__attribute__((weak))
+int nvme_submit_passthru(int fd, unsigned long ioctl_cmd,
+			 struct nvme_passthru_cmd *cmd, __u32 *result)
 {
-	printf("opcode       : %02x\n", cmd->opcode);
-	printf("flags        : %02x\n", cmd->flags);
-	printf("rsvd1        : %04x\n", cmd->rsvd1);
-	printf("nsid         : %08x\n", cmd->nsid);
-	printf("cdw2         : %08x\n", cmd->cdw2);
-	printf("cdw3         : %08x\n", cmd->cdw3);
-	printf("data_len     : %08x\n", cmd->data_len);
-	printf("metadata_len : %08x\n", cmd->metadata_len);
-	printf("addr         : %"PRIx64"\n", (uint64_t)(uintptr_t)cmd->addr);
-	printf("metadata     : %"PRIx64"\n", (uint64_t)(uintptr_t)cmd->metadata);
-	printf("cdw10        : %08x\n", cmd->cdw10);
-	printf("cdw11        : %08x\n", cmd->cdw11);
-	printf("cdw12        : %08x\n", cmd->cdw12);
-	printf("cdw13        : %08x\n", cmd->cdw13);
-	printf("cdw14        : %08x\n", cmd->cdw14);
-	printf("cdw15        : %08x\n", cmd->cdw15);
-	printf("timeout_ms   : %08x\n", cmd->timeout_ms);
-	printf("result       : %08x\n", cmd->result);
-	printf("err          : %d\n", err);
-	printf("latency      : %lu us\n",
-	       (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec));
-}
-
-void nvme_set_debug(bool debug)
-{
-	nvme_debug = debug;
-}
-
-bool nvme_get_debug(void)
-{
-	return nvme_debug;
-}
-
-static int nvme_submit_passthru(int fd, unsigned long ioctl_cmd,
-				struct nvme_passthru_cmd *cmd, __u32 *result)
-{
-	struct timeval start;
-	struct timeval end;
-	int err;
-
-	if (nvme_get_debug())
-		gettimeofday(&start, NULL);
-
-	err = ioctl(fd, ioctl_cmd, cmd);
-
-	if (nvme_get_debug()) {
-		gettimeofday(&end, NULL);
-		nvme_show_command(cmd, err, start, end);
-	}
+	int err = ioctl(fd, ioctl_cmd, cmd);
 
 	if (err >= 0 && result)
 		*result = cmd->result;
-
 	return err;
 }
 
@@ -246,125 +198,6 @@ int nvme_admin_passthru(int fd, __u8 opcode, __u8 flags, __u16 rsvd,
 			     metadata, timeout_ms, result);
 }
 
-enum nvme_cmd_dword_fields {
-	NVME_DEVICE_SELF_TEST_CDW10_STC_SHIFT			= 0,
-	NVME_DEVICE_SELF_TEST_CDW10_STC_MASK			= 0xf,
-	NVME_DIRECTIVE_CDW11_DOPER_SHIFT			= 0,
-	NVME_DIRECTIVE_CDW11_DTYPE_SHIFT			= 8,
-	NVME_DIRECTIVE_CDW11_DPSEC_SHIFT			= 16,
-	NVME_DIRECTIVE_CDW11_DOPER_MASK				= 0xff,
-	NVME_DIRECTIVE_CDW11_DTYPE_MASK				= 0xff,
-	NVME_DIRECTIVE_CDW11_DPSEC_MASK				= 0xffff,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_ENDIR_SHIFT		= 0,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_DTYPE_SHIFT		= 1,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_ENDIR_MASK		= 0x1,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_DTYPE_MASK		= 0x1,
-	NVME_FW_COMMIT_CDW10_FS_SHIFT				= 0,
-	NVME_FW_COMMIT_CDW10_CA_SHIFT				= 3,
-	NVME_FW_COMMIT_CDW10_BPID_SHIFT				= 31,
-	NVME_FW_COMMIT_CDW10_FS_MASK				= 0x7,
-	NVME_FW_COMMIT_CDW10_CA_MASK				= 0x7,
-	NVME_FW_COMMIT_CDW10_BPID_MASK				= 0x1,
-	NVME_GET_FEATURES_CDW10_SEL_SHIFT			= 8,
-	NVME_GET_FEATURES_CDW10_SEL_MASK			= 0x7,
-	NVME_SET_FEATURES_CDW10_SAVE_SHIFT			= 31,
-	NVME_SET_FEATURES_CDW10_SAVE_MASK			= 0x1,
-	NVME_FEATURES_CDW10_FID_SHIFT				= 0,
-	NVME_FEATURES_CDW14_UUID_SHIFT				= 0,
-	NVME_FEATURES_CDW10_FID_MASK				= 0xff,
-	NVME_FEATURES_CDW14_UUID_MASK				= 0x7f,
-	NVME_LOG_CDW10_LID_SHIFT				= 0,
-	NVME_LOG_CDW10_LSP_SHIFT				= 8,
-	NVME_LOG_CDW10_RAE_SHIFT				= 15,
-	NVME_LOG_CDW10_NUMDL_SHIFT				= 16,
-	NVME_LOG_CDW11_NUMDU_SHIFT				= 0,
-	NVME_LOG_CDW11_LSI_SHIFT				= 16,
-	NVME_LOG_CDW14_UUID_SHIFT				= 0,
-	NVME_LOG_CDW14_CSI_SHIFT				= 24,
-	NVME_LOG_CDW14_OT_SHIFT					= 23,
-	NVME_LOG_CDW10_LID_MASK					= 0xff,
-	NVME_LOG_CDW10_LSP_MASK					= 0x7f,
-	NVME_LOG_CDW10_RAE_MASK					= 0x1,
-	NVME_LOG_CDW10_NUMDL_MASK				= 0xffff,
-	NVME_LOG_CDW11_NUMDU_MASK				= 0xffff,
-	NVME_LOG_CDW11_LSI_MASK					= 0xffff,
-	NVME_LOG_CDW14_UUID_MASK				= 0x7f,
-	NVME_LOG_CDW14_CSI_MASK					= 0xff,
-	NVME_LOG_CDW14_OT_MASK					= 0x1,
-	NVME_IDENTIFY_CDW10_CNS_SHIFT				= 0,
-	NVME_IDENTIFY_CDW10_CNTID_SHIFT				= 16,
-	NVME_IDENTIFY_CDW11_CNSSPECID_SHIFT			= 0,
-	NVME_IDENTIFY_CDW14_UUID_SHIFT				= 0,
-	NVME_IDENTIFY_CDW11_CSI_SHIFT				= 24,
-	NVME_IDENTIFY_CDW10_CNS_MASK				= 0xff,
-	NVME_IDENTIFY_CDW10_CNTID_MASK				= 0xffff,
-	NVME_IDENTIFY_CDW11_CNSSPECID_MASK			= 0xffff,
-	NVME_IDENTIFY_CDW14_UUID_MASK				= 0x7f,
-	NVME_IDENTIFY_CDW11_CSI_MASK				= 0xff,
-	NVME_NAMESPACE_ATTACH_CDW10_SEL_SHIFT			= 0,
-	NVME_NAMESPACE_ATTACH_CDW10_SEL_MASK			= 0xf,
-	NVME_NAMESPACE_MGMT_CDW10_SEL_SHIFT			= 0,
-	NVME_NAMESPACE_MGMT_CDW10_SEL_MASK			= 0xf,
-	NVME_NAMESPACE_MGMT_CDW11_CSI_SHIFT			= 24,
-	NVME_NAMESPACE_MGMT_CDW11_CSI_MASK			= 0xff,
-	NVME_VIRT_MGMT_CDW10_ACT_SHIFT				= 0,
-	NVME_VIRT_MGMT_CDW10_RT_SHIFT				= 8,
-	NVME_VIRT_MGMT_CDW10_CNTLID_SHIFT			= 16,
-	NVME_VIRT_MGMT_CDW11_NR_SHIFT				= 0,
-	NVME_VIRT_MGMT_CDW10_ACT_MASK				= 0xf,
-	NVME_VIRT_MGMT_CDW10_RT_MASK				= 0x7,
-	NVME_VIRT_MGMT_CDW10_CNTLID_MASK			= 0xffff,
-	NVME_VIRT_MGMT_CDW11_NR_MASK				= 0xffff,
-	NVME_FORMAT_CDW10_LBAF_SHIFT				= 0,
-	NVME_FORMAT_CDW10_MSET_SHIFT				= 4,
-	NVME_FORMAT_CDW10_PI_SHIFT				= 5,
-	NVME_FORMAT_CDW10_PIL_SHIFT				= 8,
-	NVME_FORMAT_CDW10_SES_SHIFT				= 9,
-	NVME_FORMAT_CDW10_LBAFU_SHIFT				= 12,
-	NVME_FORMAT_CDW10_LBAF_MASK				= 0xf,
-	NVME_FORMAT_CDW10_MSET_MASK				= 0x1,
-	NVME_FORMAT_CDW10_PI_MASK				= 0x7,
-	NVME_FORMAT_CDW10_PIL_MASK				= 0x1,
-	NVME_FORMAT_CDW10_SES_MASK				= 0x7,
-	NVME_FORMAT_CDW10_LBAFU_MASK				= 0x3,
-	NVME_SANITIZE_CDW10_SANACT_SHIFT			= 0,
-	NVME_SANITIZE_CDW10_AUSE_SHIFT				= 3,
-	NVME_SANITIZE_CDW10_OWPASS_SHIFT			= 4,
-	NVME_SANITIZE_CDW10_OIPBP_SHIFT				= 8,
-	NVME_SANITIZE_CDW10_NODAS_SHIFT				= 9,
-	NVME_SANITIZE_CDW10_SANACT_MASK				= 0x7,
-	NVME_SANITIZE_CDW10_AUSE_MASK				= 0x1,
-	NVME_SANITIZE_CDW10_OWPASS_MASK				= 0xf,
-	NVME_SANITIZE_CDW10_OIPBP_MASK				= 0x1,
-	NVME_SANITIZE_CDW10_NODAS_MASK				= 0x1,
-	NVME_SECURITY_NSSF_SHIFT				= 0,
-	NVME_SECURITY_SPSP0_SHIFT				= 8,
-	NVME_SECURITY_SPSP1_SHIFT				= 16,
-	NVME_SECURITY_SECP_SHIFT				= 24,
-	NVME_SECURITY_NSSF_MASK					= 0xff,
-	NVME_SECURITY_SPSP0_MASK				= 0xff,
-	NVME_SECURITY_SPSP1_MASK				= 0xff,
-	NVME_SECURITY_SECP_MASK					= 0xffff,
-	NVME_GET_LBA_STATUS_CDW13_RL_SHIFT			= 0,
-	NVME_GET_LBA_STATUS_CDW13_ATYPE_SHIFT			= 24,
-	NVME_GET_LBA_STATUS_CDW13_RL_MASK			= 0xffff,
-	NVME_GET_LBA_STATUS_CDW13_ATYPE_MASK			= 0xff,
-	NVME_ZNS_MGMT_SEND_ZSASO_SHIFT				= 9,
-	NVME_ZNS_MGMT_SEND_ZSASO_MASK				= 0x1,
-	NVME_ZNS_MGMT_SEND_SEL_SHIFT				= 8,
-	NVME_ZNS_MGMT_SEND_SEL_MASK				= 0x1,
-	NVME_ZNS_MGMT_SEND_ZSA_SHIFT				= 0,
-	NVME_ZNS_MGMT_SEND_ZSA_MASK				= 0xff,
-	NVME_ZNS_MGMT_RECV_ZRA_SHIFT				= 0,
-	NVME_ZNS_MGMT_RECV_ZRA_MASK				= 0xff,
-	NVME_ZNS_MGMT_RECV_ZRASF_SHIFT				= 8,
-	NVME_ZNS_MGMT_RECV_ZRASF_MASK				= 0xff,
-	NVME_ZNS_MGMT_RECV_ZRAS_FEAT_SHIFT			= 16,
-	NVME_ZNS_MGMT_RECV_ZRAS_FEAT_MASK			= 0x1,
-	NVME_DIM_TAS_SHIFT					= 0,
-	NVME_DIM_TAS_MASK					= 0xF,
-};
-
 enum features {
 	NVME_FEATURES_ARBITRATION_BURST_SHIFT			= 0,
 	NVME_FEATURES_ARBITRATION_LPW_SHIFT			= 8,
@@ -522,6 +355,131 @@ int nvme_get_log_page(int fd, __u32 xfer_len, struct nvme_get_log_args *args)
 	return 0;
 }
 
+static int read_ana_chunk(int fd, enum nvme_log_ana_lsp lsp, bool rae,
+			  __u8 *log, __u8 **read, __u8 *to_read, __u8 *log_end)
+{
+	if (to_read > log_end) {
+		errno = ENOSPC;
+		return -1;
+	}
+
+	while (*read < to_read) {
+		__u32 len = min(log_end - *read, NVME_LOG_PAGE_PDU_SIZE);
+		int ret;
+
+		ret = nvme_get_log_ana(fd, lsp, rae, *read - log, len, *read);
+		if (ret)
+			return ret;
+
+		*read += len;
+	}
+	return 0;
+}
+
+static int try_read_ana(int fd, enum nvme_log_ana_lsp lsp, bool rae,
+			struct nvme_ana_log *log, __u8 *log_end,
+			__u8 *read, __u8 **to_read, bool *may_retry)
+{
+	__u16 ngrps = le16_to_cpu(log->ngrps);
+
+	while (ngrps--) {
+		__u8 *group = *to_read;
+		int ret;
+		__le32 nnsids;
+
+		*to_read += sizeof(*log->descs);
+		ret = read_ana_chunk(fd, lsp, rae,
+				     (__u8 *)log, &read, *to_read, log_end);
+		if (ret) {
+			/*
+			 * If the provided buffer isn't long enough,
+			 * the log page may have changed while reading it
+			 * and the computed length was inaccurate.
+			 * Have the caller check chgcnt and retry.
+			 */
+			*may_retry = errno == ENOSPC;
+			return ret;
+		}
+
+		/*
+		 * struct nvme_ana_group_desc has 8-byte alignment
+		 * but the group pointer is only 4-byte aligned.
+		 * Don't dereference the misaligned pointer.
+		 */
+		memcpy(&nnsids,
+		       group + offsetof(struct nvme_ana_group_desc, nnsids),
+		       sizeof(nnsids));
+		*to_read += le32_to_cpu(nnsids) * sizeof(__le32);
+		ret = read_ana_chunk(fd, lsp, rae,
+				     (__u8 *)log, &read, *to_read, log_end);
+		if (ret) {
+			*may_retry = errno == ENOSPC;
+			return ret;
+		}
+	}
+
+	*may_retry = true;
+	return 0;
+}
+
+int nvme_get_ana_log_atomic(int fd, bool rgo, bool rae, unsigned int retries,
+			    struct nvme_ana_log *log, __u32 *len)
+{
+	const enum nvme_log_ana_lsp lsp =
+		rgo ? NVME_LOG_ANA_LSP_RGO_GROUPS_ONLY : 0;
+	/* Get Log Page can only fetch multiples of dwords */
+	__u8 * const log_end = (__u8 *)log + (*len & -4);
+	__u8 *read = (__u8 *)log;
+	__u8 *to_read;
+	int ret;
+
+	if (!retries) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	to_read = (__u8 *)log->descs;
+	ret = read_ana_chunk(fd, lsp, rae,
+			     (__u8 *)log, &read, to_read, log_end);
+	if (ret)
+		return ret;
+
+	do {
+		bool may_retry = false;
+		int saved_ret;
+		int saved_errno;
+		__le64 chgcnt;
+
+		saved_ret = try_read_ana(fd, lsp, rae, log, log_end,
+					 read, &to_read, &may_retry);
+		/*
+		 * If the log page was read with multiple Get Log Page commands,
+		 * chgcnt must be checked afterwards to ensure atomicity
+		 */
+		*len = to_read - (__u8 *)log;
+		if (*len <= NVME_LOG_PAGE_PDU_SIZE || !may_retry)
+			return saved_ret;
+
+		saved_errno = errno;
+		chgcnt = log->chgcnt;
+		read = (__u8 *)log;
+		to_read = (__u8 *)log->descs;
+		ret = read_ana_chunk(fd, lsp, rae,
+				     (__u8 *)log, &read, to_read, log_end);
+		if (ret)
+			return ret;
+
+		if (log->chgcnt == chgcnt) {
+			/* Log hasn't changed; return try_read_ana() result */
+			errno = saved_errno;
+			return saved_ret;
+		}
+	} while (--retries);
+
+	errno = EAGAIN;
+	return -1;
+}
+
 int nvme_set_features(struct nvme_set_features_args *args)
 {
 	__u32 cdw10 = NVME_SET(args->fid, FEATURES_CDW10_FID) |
@@ -1837,35 +1795,30 @@ static int nvme_set_var_size_tags(__u32 *cmd_dw2, __u32 *cmd_dw3, __u32 *cmd_dw1
 		__u8 pif, __u8 sts, __u64 reftag, __u64 storage_tag)
 {
 	__u32 cdw2 = 0, cdw3 = 0, cdw14;
-	beint64_t be_reftag = cpu_to_be64(reftag);
-	beint64_t be_storage_tag = cpu_to_be64(storage_tag);
 
 	switch (pif) {
-	/* 16b Protection Information */
-	case 0:
-		cdw14 = be_reftag & 0xffffffff;
-		cdw14 |= ((be_storage_tag << (32 - sts)) & 0xffffffff);
+	case NVME_NVM_PIF_16B_GUARD:
+		cdw14 = reftag & 0xffffffff;
+		cdw14 |= ((storage_tag << (32 - sts)) & 0xffffffff);
 		break;
-	/* 32b Protection Information */
-	case 1:
-		cdw14 = be_reftag & 0xffffffff;
-		cdw3 = be_reftag >> 32;
-		cdw14 |= ((be_storage_tag << (80 - sts)) & 0xffff0000);
+	case NVME_NVM_PIF_32B_GUARD:
+		cdw14 = reftag & 0xffffffff;
+		cdw3 = reftag >> 32;
+		cdw14 |= ((storage_tag << (80 - sts)) & 0xffff0000);
 		if (sts >= 48)
-			cdw3 |= ((be_storage_tag >> (sts - 48)) & 0xffffffff);
+			cdw3 |= ((storage_tag >> (sts - 48)) & 0xffffffff);
 		else
-			cdw3 |= ((be_storage_tag << (48 - sts)) & 0xffffffff);
-		cdw2 = (be_storage_tag >> (sts - 16)) & 0xffff;
+			cdw3 |= ((storage_tag << (48 - sts)) & 0xffffffff);
+		cdw2 = (storage_tag >> (sts - 16)) & 0xffff;
 		break;
-	/* 64b Protection Information */
-	case 2:
-		cdw14 = be_reftag & 0xffffffff;
-		cdw3 = (be_reftag >> 32) & 0xffff;
-		cdw14 |= ((be_storage_tag << (48 - sts)) & 0xffffffff);
+	case NVME_NVM_PIF_64B_GUARD:
+		cdw14 = reftag & 0xffffffff;
+		cdw3 = (reftag >> 32) & 0xffff;
+		cdw14 |= ((storage_tag << (48 - sts)) & 0xffffffff);
 		if (sts >= 16)
-			cdw3 |= ((be_storage_tag >> (sts - 16)) & 0xffff);
+			cdw3 |= ((storage_tag >> (sts - 16)) & 0xffff);
 		else
-			cdw3 |= ((be_storage_tag << (16 - sts)) & 0xffff);
+			cdw3 |= ((storage_tag << (16 - sts)) & 0xffff);
 		break;
 	default:
 		perror("Unsupported Protection Information Format");
@@ -1873,9 +1826,9 @@ static int nvme_set_var_size_tags(__u32 *cmd_dw2, __u32 *cmd_dw3, __u32 *cmd_dw1
 		return -1;
 	}
 
-	*cmd_dw2 = cdw2;
-	*cmd_dw3 = cdw3;
-	*cmd_dw14 = cdw14;
+	*cmd_dw2 = cpu_to_be32(cdw2);
+	*cmd_dw3 = cpu_to_be32(cdw3);
+	*cmd_dw14 = cpu_to_be32(cdw14);
 	return 0;
 }
 
